@@ -1,5 +1,5 @@
 // Copyright (C) Andrew Tridgell 2002 (original file)
-// Copyright (C) 2006 Red Hat Inc. (systemtap changes)
+// Copyright (C) 2006-2007 Red Hat Inc. (systemtap changes)
 // 
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -19,6 +19,8 @@
 #include "session.h"
 #include "hash.h"
 #include "util.h"
+
+#include <cstdlib>
 #include <sstream>
 #include <iomanip>
 #include <cerrno>
@@ -63,9 +65,6 @@ hash::result(string& r)
 }
 
 
-// Grabbed from linux/module.h kernel include.
-#define MODULE_NAME_LEN (64 - sizeof(unsigned long))
-
 void
 find_hash (systemtap_session& s, const string& script)
 {
@@ -97,6 +96,7 @@ find_hash (systemtap_session& s, const string& script)
   h.add(s.bulk_mode);			// '-b'
   h.add(s.merge);			// '-M'
   h.add(s.timing);			// '-t'
+  h.add(s.prologue_searching);		// '-P'
   for (unsigned i = 0; i < s.macros.size(); i++)
     h.add(s.macros[i]);
 
