@@ -18,6 +18,7 @@
 #include "hash.h"
 #include "cache.h"
 #include "util.h"
+#include "coveragedb.h"
 
 #include <iostream>
 #include <fstream>
@@ -44,8 +45,8 @@ version ()
 {
   clog
     << "SystemTap translator/driver "
-    << "(version " << VERSION << " built " << DATE << ")" << endl
-    << "(Using " << dwfl_version (NULL) << " libraries.)" << endl
+    << "(version " << VERSION << "/" << dwfl_version (NULL)
+    << " built " << DATE << ")" << endl
     << "Copyright (C) 2005-2007 Red Hat, Inc. and others" << endl
     << "This is free software; see the source for copying conditions." << endl;
 }
@@ -70,6 +71,7 @@ usage (systemtap_session& s, int exitcode)
     << "   -V         show version" << endl
     << "   -k         keep temporary directory" << endl
     << "   -u         unoptimized translation" << (s.unoptimized ? " [set]" : "") << endl
+    << "   -w         suppress warnings" << (s.suppress_warnings ? " [set]" : "") << endl
     << "   -g         guru mode" << (s.guru_mode ? " [set]" : "") << endl
     << "   -P         prologue-searching for function probes" 
     << (s.prologue_searching ? " [set]" : "") << endl
@@ -99,6 +101,10 @@ usage (systemtap_session& s, int exitcode)
     << endl
     << "   -x PID     sets target() to PID" << endl
     << "   -t         collect probe timing information" << endl
+#ifdef HAVE_LIBSQLITE3
+    << "   -q         generate information on tapset coverage"
+#endif /* HAVE_LIBSQLITE3 */
+    << endl
     ;
   // -d: dump safety-related external references
 
@@ -186,6 +192,7 @@ main (int argc, char * const argv [])
   string script_file; // FILE
   bool have_script = false;
   bool release_changed = false;
+  bool save_module = false;
 
   // Initialize defaults
   systemtap_session s;
@@ -194,10 +201,11 @@ main (int argc, char * const argv [])
   s.kernel_release = string (buf.release);
   s.architecture = string (buf.machine);
   s.verbose = 0;
-  s.timing = 0;
+  s.timing = false;
   s.guru_mode = false;
   s.bulk_mode = false;
   s.unoptimized = false;
+  s.suppress_warnings = false;
 
 #ifdef ENABLE_PROLOGUES
   s.prologue_searching = true;
@@ -216,17 +224,17 @@ main (int argc, char * const argv [])
   s.perfmon=0;
   s.symtab = false;
   s.use_cache = true;
+  s.tapset_compile_coverage = false;
+  s.need_uprobes = false;
 
   const char* s_p = getenv ("SYSTEMTAP_TAPSET");
   if (s_p != NULL)  
   {
     s.include_path.push_back (s_p);
-    s.include_path.push_back (string(s_p) + "/LKET");
   }
   else
   {
     s.include_path.push_back (string(PKGDATADIR) + "/tapset");
-    s.include_path.push_back (string(PKGDATADIR) + "/tapset/LKET");
   }
 
   const char* s_r = getenv ("SYSTEMTAP_RUNTIME");
@@ -262,10 +270,14 @@ main (int argc, char * const argv [])
 	}
     }
 
+  const char* s_tc = getenv ("SYSTEMTAP_COVERAGE");
+  if (s_tc != NULL)
+    s.tapset_compile_coverage = true;
+
   while (true)
     {
-      // NB: also see find_hash(), help(), switch stmt below, stap.1 man page
-      int grc = getopt (argc, argv, "hVMvtp:I:e:o:R:r:m:kgPc:x:D:bs:u");
+      // NB: also see find_hash(), usage(), switch stmt below, stap.1 man page
+      int grc = getopt (argc, argv, "hVMvtp:I:e:o:R:r:m:kgPc:x:D:bs:uqw");
       if (grc < 0)
         break;
       switch (grc)
@@ -283,7 +295,11 @@ main (int argc, char * const argv [])
 	  break;
 
         case 't':
-	  s.timing ++;
+	  s.timing = true;
+	  break;
+
+        case 'w':
+	  s.suppress_warnings = true;
 	  break;
 
         case 'p':
@@ -320,6 +336,7 @@ main (int argc, char * const argv [])
 
         case 'm':
           s.module_name = string (optarg);
+	  save_module = true;
 	  {
 	    string::size_type len = s.module_name.length();
 
@@ -350,6 +367,14 @@ main (int argc, char * const argv [])
 		    " characters [_a-zA-Z0-9])." << endl;
 		usage (s, 1);
 	      }
+
+	    // Make sure module name isn't too long.
+	    if (s.module_name.size() >= (MODULE_NAME_LEN - 1))
+	      {
+		s.module_name.resize(MODULE_NAME_LEN - 1);
+		cerr << "Truncating module name to '" << s.module_name
+		     << "'" << endl;
+	      }
 	  }
 
 	  cerr << "Warning: using '-m' disables cache support." << endl;
@@ -402,6 +427,10 @@ main (int argc, char * const argv [])
 	  s.macros.push_back (string (optarg));
 	  break;
 
+	case 'q':
+	  s.tapset_compile_coverage = true;
+	  break;
+
         case 'h':
           usage (s, 0);
           break;
@@ -424,6 +453,12 @@ main (int argc, char * const argv [])
       usage (s, 1);
     }
 
+  if((s.cmd != "") && (s.target_pid))
+    {
+      cerr << "You can't specify -c and -x options together." <<endl;
+      usage (s, 1);
+    }
+
   if (s.last_pass > 4 && release_changed)
     {
       if (s.verbose)
@@ -461,6 +496,16 @@ main (int argc, char * const argv [])
            << e << endl;
     }
 
+  // Get rid of a few standard environment variables (which might
+  // cause us to do unintended things).
+  rc = unsetenv("IFS") || unsetenv("CDPATH") || unsetenv("ENV")
+      || unsetenv("BASH_ENV");
+  if (rc)
+    {
+      const char* e = strerror (errno);
+      cerr << "unsetenv failed: " << e << endl;
+    }
+
   s.kernel_base_release.assign(s.kernel_release, 0, s.kernel_release.find('-'));
 
   // arguments parsed; get down to business
@@ -563,9 +608,9 @@ main (int argc, char * const argv [])
             rc ++;
           // GLOB_NOMATCH is acceptable
 
-          if (s.verbose>1)
+          if (s.verbose>1 && globbuf.gl_pathc > 0)
             clog << "Searched '" << dir << "', "
-                 << "match count " << globbuf.gl_pathc << endl;
+                 << "found " << globbuf.gl_pathc << endl;
 
           for (unsigned j=0; j<globbuf.gl_pathc; j++)
             {
@@ -745,10 +790,25 @@ main (int argc, char * const argv [])
     cerr << "Pass 4: compilation failed.  "
          << "Try again with more '-v' (verbose) options."
          << endl;
-  else if (s.use_cache)
+  else
     {
       // Update cache.
-      add_to_cache(s);
+      if (s.use_cache)
+	add_to_cache(s);
+
+      // Copy module to the current directory.
+      if (save_module)
+        {
+	  string module_src_path = s.tmpdir + "/" + s.module_name + ".ko";
+	  string module_dest_path = s.module_name + ".ko";
+
+	  if (s.verbose > 1)
+	    clog << "Copying " << module_src_path << " to "
+		 << module_dest_path << endl;
+	  if (copy_file(module_src_path.c_str(), module_dest_path.c_str()) != 0)
+	    cerr << "Copy failed (\"" << module_src_path << "\" to \""
+		 << module_dest_path << "\"): " << strerror(errno) << endl;
+	}
     }
 
   if (rc || s.last_pass == 4) goto cleanup;
@@ -777,6 +837,16 @@ pass_5:
   // if (rc) goto cleanup;
 
  cleanup:
+
+  // update the database information
+  if (!rc && s.tapset_compile_coverage) {
+#ifdef HAVE_LIBSQLITE3
+    update_coverage_db(s);
+#else
+    cerr << "Coverage database not available without libsqlite3" << endl;
+#endif
+  }
+
   // Clean up temporary directory.  Obviously, be careful with this.
   if (s.tmpdir == "")
     ; // do nothing
