@@ -1,6 +1,7 @@
 // recursive descent parser for systemtap scripts
-// Copyright (C) 2005-2006 Red Hat Inc.
+// Copyright (C) 2005-2007 Red Hat Inc.
 // Copyright (C) 2006 Intel Corporation.
+// Copyright (C) 2007 Bull S.A.S
 //
 // This file is part of systemtap, and is free software.  You can
 // redistribute it and/or modify it under the terms of the GNU General
@@ -11,6 +12,8 @@
 #include "staptree.h"
 #include "parse.h"
 #include "session.h"
+#include "util.h"
+
 #include <iostream>
 #include <fstream>
 #include <cctype>
@@ -19,6 +22,8 @@
 #include <cerrno>
 #include <climits>
 #include <sstream>
+#include <cstring>
+#include <cctype>
 
 using namespace std;
 
@@ -146,6 +151,8 @@ parser::last ()
 // The basic form is %( CONDITION %? THEN-TOKENS %: ELSE-TOKENS %)
 // where CONDITION is: kernel_v[r] COMPARISON-OP "version-string"
 //                 or: arch COMPARISON-OP "arch-string"
+//                 or: "string1" COMPARISON-OP "string2"
+//                 or: number1 COMPARISON-OP number2
 // The %: ELSE-TOKENS part is optional.
 //
 // e.g. %( kernel_v > "2.5" %? "foo" %: "baz" %)
@@ -212,14 +219,51 @@ bool eval_pp_conditional (systemtap_session& s,
       
       return result;
     }  
+  else if ((l->type == tok_string && r->type == tok_string)
+	   || (l->type == tok_number && r->type == tok_number))
+    {
+      // collect acceptable strverscmp results.
+      int rvc_ok1, rvc_ok2;
+      if (op->type == tok_operator && op->content == "<=")
+        { rvc_ok1 = -1; rvc_ok2 = 0; }
+      else if (op->type == tok_operator && op->content == ">=")
+        { rvc_ok1 = 1; rvc_ok2 = 0; }
+      else if (op->type == tok_operator && op->content == "<")
+        { rvc_ok1 = -1; rvc_ok2 = -1; }
+      else if (op->type == tok_operator && op->content == ">")
+        { rvc_ok1 = 1; rvc_ok2 = 1; }
+      else if (op->type == tok_operator && op->content == "==")
+        { rvc_ok1 = 0; rvc_ok2 = 0; }
+      else if (op->type == tok_operator && op->content == "!=")
+        { rvc_ok1 = -1; rvc_ok2 = 1; }
+      else
+        throw parse_error ("expected comparison operator", op);
+
+      int rvc_result = l->content.compare(r->content);
+
+      // normalize rvc_result
+      if (rvc_result < 0) rvc_result = -1;
+      if (rvc_result > 0) rvc_result = 1;
+
+      return (rvc_result == rvc_ok1 || rvc_result == rvc_ok2);
+    }
+  else if (l->type == tok_string && r->type == tok_number
+	    && op->type == tok_operator)
+    throw parse_error ("expected string literal as right value", r);
+  else if (l->type == tok_number && r->type == tok_string
+	    && op->type == tok_operator)
+    throw parse_error ("expected number literal as right value", r);
   // XXX: support other forms?  "CONFIG_SMP" ?
   else
-    throw parse_error ("expected 'arch' or 'kernel_v' or 'kernel_vr'", l);
+    throw parse_error ("expected 'arch' or 'kernel_v' or 'kernel_vr'\n"
+		       "             or comparison between strings or integers", l);
 }
 
 
+// expand_args is used to know if we must expand $x and @x identifiers.
+// Only tokens corresponding to the TRUE statement must be expanded
 const token*
-parser::scan_pp ()
+parser::scan_pp (bool expand_args)
 {
   while (true)
     {
@@ -230,7 +274,7 @@ parser::scan_pp ()
           return t;
         }
 
-      const token* t = input.scan (); // NB: not recursive!
+      const token* t = input.scan (expand_args); // NB: not recursive!
       if (t == 0) // EOF
         return t;
       
@@ -240,15 +284,17 @@ parser::scan_pp ()
       // We have a %( - it's time to throw a preprocessing party!
 
       const token *l, *op, *r;
-      l = input.scan (); // NB: not recursive, though perhaps could be
-      op = input.scan ();
-      r = input.scan ();
+      l = input.scan (expand_args); // NB: not recursive, though perhaps could be
+      op = input.scan (expand_args);
+      r = input.scan (expand_args);
       if (l == 0 || op == 0 || r == 0)
         throw parse_error ("incomplete condition after '%('", t);
       // NB: consider generalizing to consume all tokens until %?, and
       // passing that as a vector to an evaluator.
 
-      bool result = eval_pp_conditional (session, l, op, r);
+      // Do not evaluate the condition if we haven't expanded everything.
+      // This may occured when having several recursive conditionals.
+      bool result = expand_args && eval_pp_conditional (session, l, op, r);
       delete l;
       delete op;
       delete r;
@@ -259,13 +305,18 @@ parser::scan_pp ()
       delete m; // "%?"
 
       vector<const token*> my_enqueued_pp;
+      bool have_token = false;
       
       while (true) // consume THEN tokens
         {
-          m = scan_pp (); // NB: recursive
+          m = scan_pp (result); // NB: recursive
           if (m == 0)
-            throw parse_error ("missing THEN tokens for conditional", t);
-          
+            throw parse_error (have_token ?
+                               "incomplete conditional - missing %: or %)" :
+                               "missing THEN tokens for conditional",
+                               t);
+
+	  have_token = true;
           if (m->type == tok_operator && (m->content == "%:" || // ELSE
                                           m->content == "%)")) // END
             break;
@@ -277,15 +328,20 @@ parser::scan_pp ()
           // continue
         }
       
+      have_token = false;
       if (m && m->type == tok_operator && m->content == "%:") // ELSE
         {
           delete m; // "%:"
           while (true)
             {
-              m = scan_pp (); // NB: recursive
+              m = scan_pp (expand_args && !result); // NB: recursive
               if (m == 0)
-                throw parse_error ("missing ELSE tokens for conditional", t);
-              
+		  throw parse_error (have_token ?
+                                     "incomplete conditional - missing %)" :
+                                     "missing ELSE tokens for conditional",
+                                     t);
+
+	      have_token = true;
               if (m->type == tok_operator && m->content == "%)") // END
                 break;
               // enqueue token
@@ -395,13 +451,37 @@ parser::expect_kw (std::string const & expected)
 }
 
 const token* 
-parser::expect_number (int64_t & expected)
+parser::expect_number (int64_t & value)
 {
-  std::string tmp;  
-  token const * tt = expect_unknown (tok_number, tmp);
-  istringstream iss(tmp);
-  iss >> expected;
-  return tt;
+  bool neg = false;
+  const token *t = next();
+  if (t->type == tok_operator && t->content == "-")
+    {
+      neg = true;
+      t = next ();
+    }
+  if (!(t && t->type == tok_number))
+    throw parse_error ("expected number");
+
+  const char* startp = t->content.c_str ();
+  char* endp = (char*) startp;
+
+  // NB: we allow controlled overflow from LLONG_MIN .. ULLONG_MAX
+  // Actually, this allows all the way from -ULLONG_MAX to ULLONG_MAX,
+  // since the lexer only gives us positive digit strings, but we'll
+  // limit it to LLONG_MIN when a '-' operator is fed into the literal.
+  errno = 0;
+  value = (int64_t) strtoull (startp, & endp, 0);
+  if (errno == ERANGE || errno == EINVAL || *endp != '\0'
+      || (neg && (unsigned long long) value > 9223372036854775808ULL)
+      || (unsigned long long) value > 18446744073709551615ULL
+      || value < -9223372036854775807LL-1)
+    throw parse_error ("number invalid or out of range"); 
+  
+  if (neg)
+    value = -value;
+
+  return t;
 }
 
 
@@ -435,7 +515,8 @@ parser::peek_kw (std::string const & kw)
 
 
 lexer::lexer (istream& i, const string& in, systemtap_session& s):
-  input (i), input_name (in), cursor_line (1), cursor_column (1), session(s)
+  input (i), input_name (in), cursor_suspend_count(0), 
+  cursor_line (1), cursor_column (1), session(s)
 { }
 
 
@@ -459,30 +540,61 @@ lexer::input_get ()
 
   if (c < 0) return c; // EOF
 
-  // update source cursor
-  if (c == '\n')
-    {
-      cursor_line ++;
-      cursor_column = 1;
-    }
+  if (cursor_suspend_count)
+    // Track effect of input_put: preserve previous cursor/line_column
+    // until all of its characters are consumed.
+    cursor_suspend_count --;
   else
-    cursor_column ++;
+    {
+      // update source cursor
+      if (c == '\n')
+        {
+          cursor_line ++;
+          cursor_column = 1;
+        }
+      else
+        cursor_column ++;
+    }
 
   return c;
 }
 
 
+void
+lexer::input_put (const string& chars)
+{
+  // clog << "[put:" << chars << "]";
+  for (int i=chars.size()-1; i>=0; i--)
+    {
+      int c = chars[i];
+      lookahead.insert (lookahead.begin(), c);
+      cursor_suspend_count ++;
+    }
+}
+
+
 token*
-lexer::scan ()
+lexer::scan (bool expand_args)
 {
   token* n = new token;
   n->location.file = input_name;
 
+  unsigned semiskipped_p = 0;
+
  skip:
   n->location.line = cursor_line;
   n->location.column = cursor_column;
 
+ semiskip:
+  if (semiskipped_p > 1)
+    {
+      input_get ();
+      throw parse_error ("invalid nested substitution of command line arguments");
+    }
+
   int c = input_get();
+  int c2 = input_peek ();
+  // clog << "{" << (char)c << (char)c2 << "}";
   if (c < 0)
     {
       delete n;
@@ -492,68 +604,77 @@ lexer::scan ()
   if (isspace (c))
     goto skip;
 
+  // Paste command line arguments as character streams into
+  // the beginning of a token.  $1..$999 go through as raw
+  // characters; @1..@999 are quoted/escaped as strings.
+  // $# and @# expand to the number of arguments, similarly
+  // raw or quoted.
+  if (expand_args &&
+      (c == '$' || c == '@') &&
+      (c2 == '#'))
+    {
+      input_get(); // swallow '#'
+      stringstream converter;
+      converter << session.args.size ();
+      if (c == '$') input_put (converter.str());
+      else input_put (lex_cast_qstring (converter.str()));
+      semiskipped_p ++;
+      goto semiskip;
+    }
+  else if (expand_args &&
+           (c == '$' || c == '@') &&
+           (isdigit (c2)))
+    {
+      unsigned idx = 0;
+      do
+        {
+          input_get ();
+          idx = (idx * 10) + (c2 - '0');
+          c2 = input_peek ();
+        } while (c2 > 0 &&
+                 isdigit (c2) && 
+                 idx <= session.args.size()); // prevent overflow
+      if (idx == 0 ||
+          idx-1 >= session.args.size())
+          throw parse_error ("command line argument index invalid or out of range", n);
+
+      string arg = session.args[idx-1];
+      if (c == '$') input_put (arg);
+      else input_put (lex_cast_qstring (arg));
+      semiskipped_p ++;
+      goto semiskip;
+    }
+
   else if (isalpha (c) || c == '$' || c == '@' || c == '_')
     {
       n->type = tok_identifier;
       n->content = (char) c;
-      while (1)
+      while (isalnum (c2) || c2 == '_' || c2 == '$')
 	{
-	  int c2 = input_peek ();
-	  if (! input)
-	    break;
-	  if ((isalnum(c2) || c2 == '_' || c2 == '$'))
-	    {
-	      n->content.push_back(c2);
-	      input_get ();
-	    }
-	  else
-	    break;
-	}
-
-      // Expand command line arguments to literals.  $1 .. $999 as
-      // numbers and @1 .. @999 as strings.
-      if (n->content[0] == '@' || n->content[0] == '$')
-        {
-          string idxstr = n->content.substr(1);
-          const char* startp = idxstr.c_str();
-          char *endp;
-          errno = 0;
-          unsigned long idx = strtoul (startp, &endp, 10);
-          if (endp == startp)
-            ; // no numbers at all - leave alone as identifier 
-          else
-            {
-              // Use @1/$1 as the base, not @0/$0.  Thus the idx-1.
-              if (errno == ERANGE || errno == EINVAL || *endp != '\0' ||
-                  idx == 0 || idx-1 >= session.args.size ())
-                throw parse_error ("command line argument index invalid or out of range");
-              
-              string arg = session.args[idx-1];
-              n->type = (n->content[0] == '@') ? tok_string : tok_number;
-              n->content = arg;
-            }
-        }
-      else
-        {
-	  if (n->content    == "probe"
-	      || n->content == "global"
-	      || n->content == "function"
-	      || n->content == "if"
-	      || n->content == "else"
-	      || n->content == "for"
-	      || n->content == "foreach"
-	      || n->content == "in"
-	      || n->content == "return"
-	      || n->content == "delete"
-	      || n->content == "while"
-	      || n->content == "break"
-	      || n->content == "continue"
-	      || n->content == "next"
-	      || n->content == "string"
-	      || n->content == "long")
-	    n->type = tok_keyword;
+          input_get ();
+          n->content.push_back (c2);
+          c2 = input_peek ();
         }
 
+      if (n->content    == "probe"
+          || n->content == "global"
+          || n->content == "function"
+          || n->content == "if"
+          || n->content == "else"
+          || n->content == "for"
+          || n->content == "foreach"
+          || n->content == "in"
+          || n->content == "limit"
+          || n->content == "return"
+          || n->content == "delete"
+          || n->content == "while"
+          || n->content == "break"
+          || n->content == "continue"
+          || n->content == "next"
+          || n->content == "string"
+          || n->content == "long")
+        n->type = tok_keyword;
+      
       return n;
     }
 
@@ -565,7 +686,7 @@ lexer::scan ()
       while (1)
 	{
 	  int c2 = input_peek ();
-	  if (! input)
+	  if (c2 < 0)
 	    break;
 
           // NB: isalnum is very permissive.  We rely on strtol, called in
@@ -590,14 +711,14 @@ lexer::scan ()
 	{
 	  c = input_get ();
 
-	  if (! input || c == '\n')
+	  if (c < 0 || c == '\n')
 	    {
 	      n->type = tok_junk;
 	      break;
 	    }
 	  if (c == '\"') // closing double-quotes
 	    break;
-	  else if (c == '\\')
+	  else if (c == '\\') // see also input_put
 	    {	      
 	      c = input_get ();
 	      switch (c)
@@ -611,14 +732,13 @@ lexer::scan ()
 		case 'r':
 		case '0' ... '7': // NB: need only match the first digit
 		case '\\':
-
 		  // Pass these escapes through to the string value
-		  // beign parsed; it will be emitted into a C literal. 
+		  // being parsed; it will be emitted into a C literal. 
 
 		  n->content.push_back('\\');
 
+                  // fall through
 		default:
-
 		  n->content.push_back(c);
 		  break;
 		}
@@ -802,24 +922,25 @@ parser::parse ()
       catch (parse_error& pe)
 	{
 	  print_error (pe);
-          try 
-            {
-              // Quietly swallow all tokens until the next '}'.
-              while (1)
-                {
-                  const token* t = peek ();
-                  if (! t)
-                    break;
-                  next ();
-                  if (t->type == tok_operator && t->content == "}")
-                    break;
-                }
-            }
-          catch (parse_error& pe2)
-            {
-              // parse error during recovery ... ugh
-              print_error (pe2);
-            }
+          if (pe.skip_some) // for recovery
+            try 
+              {
+                // Quietly swallow all tokens until the next '}'.
+                while (1)
+                  {
+                    const token* t = peek ();
+                    if (! t)
+                      break;
+                    next ();
+                    if (t->type == tok_operator && t->content == "}")
+                      break;
+                  }
+              }
+            catch (parse_error& pe2)
+              {
+                // parse error during recovery ... ugh
+                print_error (pe2);
+              }
         }
     }
 
@@ -875,7 +996,6 @@ parser::parse_probe (std::vector<probe *> & probe_ret,
           next ();
           continue;
         }
-
       else if (t && t->type == tok_operator && t->content == ",")
         {
           locations.push_back(pp);
@@ -898,6 +1018,7 @@ parser::parse_probe (std::vector<probe *> & probe_ret,
       p->tok = t0;
       p->locations = locations;
       p->body = parse_stmt_block ();
+      p->privileged = privileged;
       probe_ret.push_back (p);
     }
   else
@@ -910,6 +1031,7 @@ parser::parse_probe (std::vector<probe *> & probe_ret,
       p->tok = t0;
       p->locations = locations;
       p->body = parse_stmt_block ();
+      p->privileged = privileged;
       alias_ret.push_back (p);
     }
 }
@@ -924,7 +1046,8 @@ parser::parse_embeddedcode ()
     throw parse_error ("expected '%{'");
 
   if (! privileged)
-    throw parse_error ("embedded code in unprivileged script");
+    throw parse_error ("embedded code in unprivileged script",
+                       false /* don't skip tokens for parse resumption */);
 
   e->tok = t;
   e->code = t->content;
@@ -1020,7 +1143,7 @@ parser::parse_statement ()
 
 
 void
-parser::parse_global (vector <vardecl*>& globals, vector<probe*>& probes)
+parser::parse_global (vector <vardecl*>& globals, vector<probe*>&)
 {
   const token* t0 = next ();
   if (! (t0->type == tok_keyword && t0->content == "global"))
@@ -1042,54 +1165,29 @@ parser::parse_global (vector <vardecl*>& globals, vector<probe*>& probes)
       globals.push_back (d);
 
       t = peek ();
+
+      if (t && t->type == tok_operator && t->content == "[") // array size
+	{
+	  int64_t size;
+	  next ();
+	  expect_number(size);
+	  if (size <= 0 || size > 1000000) // arbitrary max
+	    throw parse_error("array size out of range");
+	  d->maxsize = (int)size;
+	  expect_known(tok_operator, "]");
+	  t = peek ();
+	}
+
       if (t && t->type == tok_operator && t->content == "=") // initialization
-        {
-          next ();
-
-          // Create synthetic "begin" probe to assign value to global.
-          //
-          // An alternative would be to store the initializer value
-          // within a new field of vardecl, and use e.g. the
-          // type-resolution stage to do the rewriting.  It could go
-          // farther: the initializer could live through till
-          // translation and be emitted as a plain C-level variable
-          // initializer.
-
-          literal* value = parse_literal ();
-
-          probe_point* pp = new probe_point;
-          probe_point::component* ppc = new probe_point::component;
-          ppc->functor = string("begin");
-          pp->tok = t;
-          pp->components.push_back (ppc);
-
-          probe* p = new probe;
-          p->tok = t;
-          p->locations.push_back (pp);
-
-          symbol* sym = new symbol;
-          sym->tok = t;
-          sym->name = d->name;
-
-          assignment* a = new assignment;
-          a->tok = t;
-          a->op = "=";
-          a->left = sym;
-          a->right = value;
-
-          expr_statement* es = new expr_statement;
-          es->tok = t;
-          es->value = a;
-
-          block* blk = new block;
-          blk->tok = t;
-          blk->statements.push_back (es);
-
-          p->body = blk;
-          probes.push_back (p);
-
-          t = peek ();
-        }
+	{
+	  if (!d->compatible_arity(0))
+	    throw parse_error("only scalar globals can be initialized");
+	  d->set_arity(0);
+	  next ();
+	  d->init = parse_literal ();
+	  d->type = d->init->type;
+	  t = peek ();
+	}
 
       if (t && t->type == tok_operator && t->content == ",") // next global
 	{
@@ -1249,9 +1347,30 @@ parser::parse_probe_point ()
       // We only fall through here at the end of a probe point (past
       // all the dotted/parametrized components).
 
-      if (t && t->type == tok_operator && t->content == "?")
+      if (t && t->type == tok_operator &&
+          (t->content == "?" || t->content == "!"))
         {
           pl->optional = true;
+          if (t->content == "!") pl->sufficient = true;
+          // NB: sufficient implies optional
+          next ();
+          t = peek ();
+          // fall through
+        }
+
+      if (t && t->type == tok_keyword && t->content == "if")
+        {
+          next ();
+          t = peek ();
+          if (! (t->type == tok_operator && t->content == "("))
+            throw parse_error ("expected '('");
+          next ();
+
+          pl->condition = parse_expression ();
+
+          t = peek ();
+          if (! (t->type == tok_operator && t->content == ")"))
+            throw parse_error ("expected ')'");
           next ();
           t = peek ();
           // fall through
@@ -1262,7 +1381,7 @@ parser::parse_probe_point ()
               t->content == "=" || t->content == "+=" ))
         break;
       
-      throw parse_error ("expected '.' or ',' or '(' or '?' or '{' or '=' or '+='");
+      throw parse_error ("expected one of '. , ( ? ! { = +='");
     }
 
   return pl;
@@ -1276,25 +1395,40 @@ parser::parse_literal ()
   literal* l;
   if (t->type == tok_string)
     l = new literal_string (t->content);
-  else if (t->type == tok_number)
-    {
-      const char* startp = t->content.c_str ();
-      char* endp = (char*) startp;
-
-      // NB: we allow controlled overflow from LLONG_MIN .. ULLONG_MAX
-      // Actually, this allows all the way from -ULLONG_MAX to ULLONG_MAX,
-      // since the lexer only gives us positive digit strings.
-      errno = 0;
-      long long value = (long long) strtoull (startp, & endp, 0);
-      if (errno == ERANGE || errno == EINVAL || *endp != '\0'
-          || (unsigned long long) value > 18446744073709551615ULL
-          || value < -9223372036854775807LL-1)
-        throw parse_error ("number invalid or out of range"); 
-
-      l = new literal_number (value);
-    }
   else
-    throw parse_error ("expected literal string or number");
+    {
+      bool neg = false;
+      if (t->type == tok_operator && t->content == "-")
+	{
+	  neg = true;
+	  t = next ();
+	}
+
+      if (t->type == tok_number)
+	{
+	  const char* startp = t->content.c_str ();
+	  char* endp = (char*) startp;
+
+	  // NB: we allow controlled overflow from LLONG_MIN .. ULLONG_MAX
+	  // Actually, this allows all the way from -ULLONG_MAX to ULLONG_MAX,
+	  // since the lexer only gives us positive digit strings, but we'll
+	  // limit it to LLONG_MIN when a '-' operator is fed into the literal.
+	  errno = 0;
+	  long long value = (long long) strtoull (startp, & endp, 0);
+	  if (errno == ERANGE || errno == EINVAL || *endp != '\0'
+	      || (neg && (unsigned long long) value > 9223372036854775808ULL)
+	      || (unsigned long long) value > 18446744073709551615ULL
+	      || value < -9223372036854775807LL-1)
+	    throw parse_error ("number invalid or out of range"); 
+
+	  if (neg)
+	    value = -value;
+
+	  l = new literal_number (value);
+	}
+      else
+	throw parse_error ("expected literal string or number");
+    }
 
   l->tok = t;
   return l;
@@ -1518,6 +1652,7 @@ parser::parse_foreach_loop ()
   foreach_loop* s = new foreach_loop;
   s->tok = t;
   s->sort_direction = 0;
+  s->limit = NULL;
 
   t = next ();
   if (! (t->type == tok_operator && t->content == "("))
@@ -1591,6 +1726,13 @@ parser::parse_foreach_loop ()
       next();
     }
 
+  t = peek ();
+  if (tok_is(t, tok_keyword, "limit"))
+    {
+      next ();				// get past the "limit"
+      s->limit = parse_expression ();
+    }
+
   t = next ();
   if (! (t->type == tok_operator && t->content == ")"))
     throw parse_error ("expected ')'");
@@ -2073,22 +2215,6 @@ parser::parse_hist_op_or_bare_name (hist_op *&hop, string &name)
 	      hop->params.push_back (tnum);
 	    }
 	}
-      else
-	{
-	  assert(hop->htype == hist_log);
-	  if (peek_op (","))
-	    {
-	      expect_op (",");
-	      expect_number (tnum);
-	      hop->params.push_back (tnum);
-	    }
-	  else
-	    {
-	      // FIXME (magic value): Logarithmic histograms get 64
-	      // buckets by default.
-	      hop->params.push_back (64);
-	    }	      
-	}
       expect_op(")");
     }
   return t;
@@ -2128,6 +2254,8 @@ parser::parse_symbol ()
       // now scrutinize this identifier for the various magic forms of identifier
       // (printf, @stat_op, and $var...)
 
+      bool pf_stream, pf_format, pf_delim, pf_newline, pf_char;
+
       if (name.size() > 0 && name[0] == '@')
 	{
 	  stat_op *sop = new stat_op;
@@ -2150,36 +2278,20 @@ parser::parse_symbol ()
 	  return sop;
 	}
       
-      else if (name.size() > 0 && (name == "print"
-				   || name == "sprint"
-				   || name == "printf"
-				   || name == "sprintf"))
+      else if (print_format::parse_print(name,
+	 pf_stream, pf_format, pf_delim, pf_newline, pf_char))
 	{
 	  print_format *fmt = new print_format;
 	  fmt->tok = t;
-	  fmt->print_with_format = (name[name.size() - 1] == 'f');
-	  fmt->print_to_stream = (name[0] == 'p');
+	  fmt->print_to_stream = pf_stream;
+	  fmt->print_with_format = pf_format;
+	  fmt->print_with_delim = pf_delim;
+	  fmt->print_with_newline = pf_newline;
+	  fmt->print_char = pf_char;
 
 	  expect_op("(");
-	  if (fmt->print_with_format)
-	    {
-	      // Consume and convert a format string, and any subsequent
-	      // arguments. Agreement between the format string and the
-	      // arguments is postponed to the typechecking phase. 
-	      string tmp;
-	      expect_unknown (tok_string, tmp);
-              fmt->raw_components = tmp;
-	      fmt->components = print_format::string_to_components (tmp);
-	      while (!peek_op (")"))
-		{
-		  expect_op(",");
-		  expression *e = parse_expression ();	      
-		  fmt->args.push_back(e);
-		}
-	    }
-	  else if (name == "print" && 
-		   (peek_kw("@hist_linear") || 
-		    peek_kw("@hist_log")))
+	  if ((name == "print" || name == "println") &&
+	      (peek_kw("@hist_linear") || peek_kw("@hist_log")))
 	    {
 	      // We have a special case where we recognize
 	      // print(@hist_foo(bar)) as a magic print-the-histogram
@@ -2214,10 +2326,42 @@ parser::parse_symbol ()
 	    }
 	  else
 	    {
-	      // If we are not printing with a format string, we permit
-	      // exactly one argument (of any type).
-	      expression *e = parse_expression ();	      
-	      fmt->args.push_back(e);
+	      int min_args = 0;
+	      if (fmt->print_with_format)
+		{
+		  // Consume and convert a format string. Agreement between the
+		  // format string and the arguments is postponed to the
+		  // typechecking phase.
+		  string tmp;
+		  expect_unknown (tok_string, tmp);
+		  fmt->raw_components = tmp;
+		  fmt->components = print_format::string_to_components (tmp);
+		}
+	      else if (fmt->print_with_delim)
+		{
+		  // Consume a delimiter to separate arguments.
+		  fmt->delimiter.clear();
+		  fmt->delimiter.type = print_format::conv_literal;
+		  expect_unknown (tok_string, fmt->delimiter.literal_string);
+		  min_args = 2;
+		}
+	      else
+		{
+		  // If we are not printing with a format string, we must have
+		  // at least one argument (of any type).
+		  expression *e = parse_expression ();
+		  fmt->args.push_back(e);
+		}
+
+	      // Consume any subsequent arguments.
+	      while (min_args || !peek_op (")"))
+		{
+		  expect_op(",");
+		  expression *e = parse_expression ();
+		  fmt->args.push_back(e);
+		  if (min_args)
+		    --min_args;
+		}
 	    }
 	  expect_op(")");
 	  return fmt;
