@@ -1,6 +1,7 @@
 /* target operations
  * Copyright (C) 2005 Red Hat Inc.
  * Copyright (C) 2005, 2006, 2007 Intel Corporation.
+ * Copyright (C) 2007 Quentin Barnes.
  *
  * This file is part of systemtap, and is free software.  You can
  * redistribute it and/or modify it under the terms of the GNU General
@@ -59,6 +60,21 @@
    must work right for kernel addresses, and can use whatever existing
    machine-specific kernel macros are convenient.  */
 
+#define DEREF_FAULT(addr) ({						    \
+    snprintf(c->error_buffer, sizeof(c->error_buffer),			    \
+      "kernel read fault at 0x%p (%s)", (void *)(intptr_t)(addr), #addr);   \
+    c->last_error = c->error_buffer;					    \
+    goto deref_fault;							    \
+    })
+
+#define STORE_DEREF_FAULT(addr) ({					    \
+    snprintf(c->error_buffer, sizeof(c->error_buffer),			    \
+      "kernel write fault at 0x%p (%s)", (void *)(intptr_t)(addr), #addr);  \
+    c->last_error = c->error_buffer;					    \
+    goto deref_fault;							    \
+    })
+
+
 #if defined __i386__
 
 /* The stack pointer is unlike other registers.  When a trap happens in
@@ -84,7 +100,7 @@
 #undef store_register
 
 #define fetch_register(regno)		ia64_fetch_register(regno, c->regs)
-#define store_register(regno, value)	ia64_store_register(regno, c->regs, value)
+#define store_register(regno,value)	ia64_store_register(regno, c->regs, value)
 
 #elif defined __x86_64__
 
@@ -110,20 +126,20 @@
 #undef fetch_register
 #undef store_register
 #define fetch_register(regno) ((intptr_t) c->regs->gpr[regno])
-#define store_register(regno) (c->regs->gpr[regno] = (value))
+#define store_register(regno,value) (c->regs->gpr[regno] = (value))
 
 #elif defined (__arm__)
 
 #undef fetch_register
 #undef store_register
 #define fetch_register(regno) ((long) c->regs->uregs[regno])
-#define store_register(regno) (c->regs->uregs[regno] = (value))
+#define store_register(regno,value) (c->regs->uregs[regno] = (value))
 
 #elif defined (__s390__) || defined (__s390x__)
 #undef fetch_register
 #undef store_register
 #define fetch_register(regno) ((intptr_t) c->regs->gprs[regno])
-#define store_register(regno) (c->regs->gprs[regno] = (value))
+#define store_register(regno,value) (c->regs->gprs[regno] = (value))
 
 #endif
 
@@ -142,7 +158,7 @@
       default: _v = __get_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      DEREF_FAULT(addr);							      \
     _v;									      \
   })
 
@@ -157,7 +173,7 @@
       default: __put_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      STORE_DEREF_FAULT(addr);							      \
   })
 
 
@@ -177,7 +193,7 @@
       default: _v = __get_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      DEREF_FAULT(addr);							      \
     _v;									      \
   })
 
@@ -193,7 +209,7 @@
       default: __put_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      STORE_DEREF_FAULT(addr);							      \
   })
 
 #elif defined __ia64__
@@ -209,7 +225,7 @@
 	default: __get_user_unknown(); break;				\
 	}								\
     if (_bad)  								\
-	goto deref_fault;						\
+	DEREF_FAULT(addr);						\
      _v;								\
    })
 
@@ -224,7 +240,7 @@
 	default: __put_user_unknown(); break;				\
 	}								\
     if (_bad)								\
-	   goto deref_fault;						\
+	   STORE_DEREF_FAULT(addr);						\
    })
 
 #elif defined __powerpc__ || defined __powerpc64__
@@ -282,7 +298,7 @@
       default: _v = __get_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      DEREF_FAULT(addr);							      \
     _v;									      \
   })
 
@@ -298,23 +314,157 @@
       default: __put_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      STORE_DEREF_FAULT(addr);							      \
   })
 
 #elif defined (__arm__)
 
+/* Macros for ARM lifted from 2.6.21.1's linux/include/asm-arm/uaccess.h
+ * and slightly altered. */
+
+#define __stp_get_user_asm_byte(x,addr,err)			\
+	__asm__ __volatile__(					\
+	"1:	ldrb	%1,[%2],#0\n"				\
+	"2:\n"							\
+	"	.section .fixup,\"ax\"\n"			\
+	"	.align	2\n"					\
+	"3:	mov	%0, %3\n"				\
+	"	mov	%1, #0\n"				\
+	"	b	2b\n"					\
+	"	.previous\n"					\
+	"	.section __ex_table,\"a\"\n"			\
+	"	.align	3\n"					\
+	"	.long	1b, 3b\n"				\
+	"	.previous"					\
+	: "+r" (err), "=&r" (x)					\
+	: "r" (addr), "i" (-EFAULT)				\
+	: "cc")
+
+#ifndef __ARMEB__
+#define __stp_get_user_asm_half(x,__gu_addr,err)		\
+({								\
+	unsigned long __b1, __b2;				\
+	__stp_get_user_asm_byte(__b1, __gu_addr, err);		\
+	__stp_get_user_asm_byte(__b2, __gu_addr + 1, err);	\
+	(x) = __b1 | (__b2 << 8);				\
+})
+#else
+#define __stp_get_user_asm_half(x,__gu_addr,err)		\
+({								\
+	unsigned long __b1, __b2;				\
+	__stp_get_user_asm_byte(__b1, __gu_addr, err);		\
+	__stp_get_user_asm_byte(__b2, __gu_addr + 1, err);	\
+	(x) = (__b1 << 8) | __b2;				\
+})
+#endif
+
+#define __stp_get_user_asm_word(x,addr,err)			\
+	__asm__ __volatile__(					\
+	"1:	ldr	%1,[%2],#0\n"				\
+	"2:\n"							\
+	"	.section .fixup,\"ax\"\n"			\
+	"	.align	2\n"					\
+	"3:	mov	%0, %3\n"				\
+	"	mov	%1, #0\n"				\
+	"	b	2b\n"					\
+	"	.previous\n"					\
+	"	.section __ex_table,\"a\"\n"			\
+	"	.align	3\n"					\
+	"	.long	1b, 3b\n"				\
+	"	.previous"					\
+	: "+r" (err), "=&r" (x)					\
+	: "r" (addr), "i" (-EFAULT)				\
+	: "cc")
+
+#define __stp_put_user_asm_byte(x,__pu_addr,err)		\
+	__asm__ __volatile__(					\
+	"1:	strb	%1,[%2],#0\n"				\
+	"2:\n"							\
+	"	.section .fixup,\"ax\"\n"			\
+	"	.align	2\n"					\
+	"3:	mov	%0, %3\n"				\
+	"	b	2b\n"					\
+	"	.previous\n"					\
+	"	.section __ex_table,\"a\"\n"			\
+	"	.align	3\n"					\
+	"	.long	1b, 3b\n"				\
+	"	.previous"					\
+	: "+r" (err)						\
+	: "r" (x), "r" (__pu_addr), "i" (-EFAULT)		\
+	: "cc")
+
+#ifndef __ARMEB__
+#define __stp_put_user_asm_half(x,__pu_addr,err)			\
+({									\
+	unsigned long __temp = (unsigned long)(x);			\
+	__stp_put_user_asm_byte(__temp, __pu_addr, err);		\
+	__stp_put_user_asm_byte(__temp >> 8, __pu_addr + 1, err);	\
+})
+#else
+#define __stp_put_user_asm_half(x,__pu_addr,err)			\
+({									\
+	unsigned long __temp = (unsigned long)(x);			\
+	__stp_put_user_asm_byte(__temp >> 8, __pu_addr, err);		\
+	__stp_put_user_asm_byte(__temp, __pu_addr + 1, err);		\
+})
+#endif
+
+#define __stp_put_user_asm_word(x,__pu_addr,err)		\
+	__asm__ __volatile__(					\
+	"1:	str	%1,[%2],#0\n"				\
+	"2:\n"							\
+	"	.section .fixup,\"ax\"\n"			\
+	"	.align	2\n"					\
+	"3:	mov	%0, %3\n"				\
+	"	b	2b\n"					\
+	"	.previous\n"					\
+	"	.section __ex_table,\"a\"\n"			\
+	"	.align	3\n"					\
+	"	.long	1b, 3b\n"				\
+	"	.previous"					\
+	: "+r" (err)						\
+	: "r" (x), "r" (__pu_addr), "i" (-EFAULT)		\
+	: "cc")
+
+#ifndef __ARMEB__
+#define	__reg_oper0	"%R2"
+#define	__reg_oper1	"%Q2"
+#else
+#define	__reg_oper0	"%Q2"
+#define	__reg_oper1	"%R2"
+#endif
+
+#define __stp_put_user_asm_dword(x,__pu_addr,err)		\
+	__asm__ __volatile__(					\
+	"1:	str	" __reg_oper1 ", [%1], #4\n"		\
+	"2:	str	" __reg_oper0 ", [%1], #0\n"		\
+	"3:\n"							\
+	"	.section .fixup,\"ax\"\n"			\
+	"	.align	2\n"					\
+	"4:	mov	%0, %3\n"				\
+	"	b	3b\n"					\
+	"	.previous\n"					\
+	"	.section __ex_table,\"a\"\n"			\
+	"	.align	3\n"					\
+	"	.long	1b, 4b\n"				\
+	"	.long	2b, 4b\n"				\
+	"	.previous"					\
+	: "+r" (err), "+r" (__pu_addr)				\
+	: "r" (x), "i" (-EFAULT)				\
+	: "cc")
+
 #define deref(size, addr)						\
   ({									\
      int _bad = 0;							\
      intptr_t _v=0;							\
 	switch (size){							\
-	case 1: __get_user_asm_byte(_v, addr, _bad); break; 		\
-	case 2: __get_user_asm_half(_v, addr, _bad); break;  		\
-	case 4: __get_user_asm_word(_v, addr, _bad); break;  		\
+	case 1: __stp_get_user_asm_byte(_v, addr, _bad); break; 	\
+	case 2: __stp_get_user_asm_half(_v, addr, _bad); break; 	\
+	case 4: __stp_get_user_asm_word(_v, addr, _bad); break; 	\
 	default: __get_user_bad(); break;				\
 	}								\
     if (_bad)  								\
-	goto deref_fault;						\
+	DEREF_FAULT(addr);						\
      _v;								\
    })
 
@@ -322,14 +472,14 @@
   ({									\
     int _bad=0;								\
 	switch (size){							\
-	case 1: __put_user_asm_byte(value, addr, _bad); break;		\
-	case 2: __put_user_asm_half(value, addr, _bad); break;		\
-	case 4: __put_user_asm_word(value, addr, _bad); break;		\
-	case 8: __put_user_asm_dword(value, addr, _bad); break;		\
+	case 1: __stp_put_user_asm_byte(value, addr, _bad); break;	\
+	case 2: __stp_put_user_asm_half(value, addr, _bad); break;	\
+	case 4: __stp_put_user_asm_word(value, addr, _bad); break;	\
+	case 8: __stp_put_user_asm_dword(value, addr, _bad); break;	\
 	default: __put_user_bad(); break;				\
 	}								\
     if (_bad)								\
-	   goto deref_fault;						\
+	   STORE_DEREF_FAULT(addr);						\
    })
 
 #elif defined (__s390__) || defined (__s390x__)
@@ -415,7 +565,7 @@
 		_bad = -EFAULT;					\
 	}							\
 	if (_bad)						\
-		goto deref_fault;				\
+		DEREF_FAULT(addr);				\
 	_v;							\
 })
 
@@ -427,7 +577,7 @@
 		__stp_put_asm((u8)(value>>((size-i-1)*8)&0xff),	\
                              (u64)addr+i,_bad); 		\
 		if (_bad)					\
-			goto deref_fault;			\
+			STORE_DEREF_FAULT(addr);			\
         }                                                       \
 })
 
@@ -482,6 +632,5 @@
 
 #define CATCH_DEREF_FAULT()				\
   if (0) {						\
-deref_fault:						\
-    CONTEXT->last_error = "pointer dereference fault";	\
+deref_fault: ;						\
   }
