@@ -47,6 +47,8 @@ unsigned long _stp_ret_addr (struct pt_regs *regs)
 	return regs->b0;
 #elif defined (__s390__) || defined (__s390x__)
 	return regs->gprs[14];
+#elif defined (__arm__)
+	return regs->ARM_r0;
 #else
 	#error Unimplemented architecture
 #endif
@@ -251,7 +253,7 @@ void _stp_print_regs(struct pt_regs * regs)
 		processor_modes[processor_mode(regs)],
 		thumb_mode(regs) ? " (T)" : "",
 		get_fs() == get_ds() ? "kernel" : "user");
-#if CONFIG_CPU_CP15
+#ifdef CONFIG_CPU_CP15
 	{
 		unsigned int ctrl;
 		  __asm__ (
