@@ -7,6 +7,15 @@
  * Public License (GPL); either version 2, or (at your option) any
  * later version.
  */
+#if defined (__i386__) || defined (__x86_64__)
+#include <asm/cpufeature.h>
+#endif
+#ifdef STAPCONF_TSC_KHZ
+#include <asm/tsc.h>
+#endif
+#ifdef STAPCONF_KTIME_GET_REAL
+#include <linux/ktime.h>
+#endif
 
 #include <linux/cpufreq.h>
 
@@ -17,10 +26,9 @@
 typedef struct __stp_time_t {
     /* 
      * A write lock is taken by __stp_time_timer_callback() and
-     * __stp_time_cpufreq_callback().  The timer callback is called from a
-     * softIRQ, and cpufreq callback guarantees that it is not called within
-     * an interrupt context.  Thus there should be no opportunity for a
-     * deadlock between writers.
+     * __stp_time_cpufreq_callback().  Neither writer is in interrupt context,
+     * and both disable interrupts before taking the lock, so there should be
+     * no opportunity for deadlock.
      *
      * A read lock is taken by _stp_gettimeofday_us().  There is the potential
      * for this to occur at any time, so there is a slim chance that this will
@@ -28,8 +36,9 @@ typedef struct __stp_time_t {
      * read lock.  However, we can limit how long we try to get the lock to
      * avoid a deadlock.
      *
-     * Note that seqlock is safer than rwlock because some kernels
-     * don't have read_trylock.
+     * Note that seqlock is chosen so that readers don't block writers.  It's
+     * also important that readers can attempt a lock from _any_ context (e.g.,
+     * NMI), and some kernels don't have read_trylock.
      */
     seqlock_t lock;
 
@@ -37,9 +46,9 @@ typedef struct __stp_time_t {
     int64_t base_ns;
     cycles_t base_cycles;
 
-    /* The frequency in kHz of this CPU, for interpolating
+    /* The frequency in kHz of this CPU's time stamp counter, for interpolating
      * cycle counts from the base time. */
-    unsigned int cpufreq;
+    unsigned int freq;
 
     /* Callback used to schedule updates of the base time */
     struct timer_list timer;
@@ -47,6 +56,9 @@ typedef struct __stp_time_t {
 
 void *stp_time = NULL;
 
+/* Flag to tell the timer callback whether to reregister */
+int stp_timer_reregister = 0;
+
 /* Try to estimate the number of CPU cycles in a millisecond - i.e. kHz.  This
  * relies heavily on the accuracy of udelay.  By calling udelay twice, we
  * attempt to account for overhead in the call.
@@ -56,14 +68,23 @@ void *stp_time = NULL;
  * FIXME: This not very accurate on Xen kernels!
  */
 static unsigned int
-__stp_estimate_cpufreq(void)
+__stp_get_freq(void)
 {
-#if defined (__s390__) || defined (__s390x__)
+    // If we can get the actual frequency of HW counter, we use it.
+#if defined (__i386__) || defined (__x86_64__)
+#ifdef STAPCONF_TSC_KHZ
+    return tsc_khz;
+#else /*STAPCONF_TSC_KHZ*/
+    return cpu_khz;
+#endif /*STAPCONF_TSC_KHZ*/
+#elif defined (__ia64__)
+    return local_cpu_data->itc_freq / 1000;
+#elif defined (__s390__) || defined (__s390x__) || defined (__arm__)
     // We don't need to find the cpu freq on s390 as the 
     // TOD clock is always a fix freq. (see: POO pg 4-36.)
     return 0;
-#else /* __s390x__ || __s390x__ */
-
+#else /* __i386__ || __x86_64__ || __ia64__ || __s390__ || __s390x__ || __arm__*/
+    // If we don't know the actual frequency, we estimate it.
     cycles_t beg, mid, end;
     beg = get_cycles(); barrier();
     udelay(1); barrier();
@@ -74,22 +95,35 @@ __stp_estimate_cpufreq(void)
 #endif
 }
 
+static void
+__stp_ktime_get_real_ts(struct timespec *ts)
+{
+#ifdef STAPCONF_KTIME_GET_REAL
+    ktime_get_real_ts(ts);
+#else /* STAPCONF_KTIME_GET_REAL */
+    struct timeval tv;
+    do_gettimeofday(&tv);
+    ts->tv_sec = tv.tv_sec;
+    ts->tv_nsec = tv.tv_usec * NSEC_PER_USEC;
+#endif
+}
+
+/* The timer callback is in a softIRQ -- interrupts enabled. */
 static void
 __stp_time_timer_callback(unsigned long val)
 {
     unsigned long flags;
     stp_time_t *time;
-    struct timeval tv;
+    struct timespec ts;
     int64_t ns;
     cycles_t cycles;
 
     local_irq_save(flags);
 
-    do_gettimeofday(&tv);
+    __stp_ktime_get_real_ts(&ts);
     cycles = get_cycles();
 
-    ns = (NSEC_PER_SEC * (int64_t)tv.tv_sec)
-        + (NSEC_PER_USEC * tv.tv_usec);
+    ns = (NSEC_PER_SEC * (int64_t)ts.tv_sec) + ts.tv_nsec;
 
     time = per_cpu_ptr(stp_time, smp_processor_id());
     write_seqlock(&time->lock);
@@ -97,7 +131,8 @@ __stp_time_timer_callback(unsigned long val)
     time->base_cycles = cycles;
     write_sequnlock(&time->lock);
 
-    mod_timer(&time->timer, jiffies + 1);
+    if (likely(stp_timer_reregister))
+        mod_timer(&time->timer, jiffies + 1);
 
     local_irq_restore(flags);
 }
@@ -106,15 +141,14 @@ __stp_time_timer_callback(unsigned long val)
 static void
 __stp_init_time(void *info)
 {
-    struct timeval tv;
+    struct timespec ts;
     stp_time_t *time = per_cpu_ptr(stp_time, smp_processor_id());
 
     seqlock_init(&time->lock);
-    do_gettimeofday(&tv);
+    __stp_ktime_get_real_ts(&ts);
     time->base_cycles = get_cycles();
-    time->base_ns = (NSEC_PER_SEC * (int64_t)tv.tv_sec)
-        + (NSEC_PER_USEC * tv.tv_usec);
-    time->cpufreq = __stp_estimate_cpufreq();
+    time->base_ns = (NSEC_PER_SEC * (int64_t)ts.tv_sec) + ts.tv_nsec;
+    time->freq = __stp_get_freq();
 
     init_timer(&time->timer);
     time->timer.expires = jiffies + 1;
@@ -123,6 +157,7 @@ __stp_init_time(void *info)
 }
 
 #ifdef CONFIG_CPU_FREQ
+/* The cpufreq callback is not in interrupt context -- interrupts enabled */
 static int
 __stp_time_cpufreq_callback(struct notifier_block *self,
         unsigned long state, void *vfreqs)
@@ -137,9 +172,9 @@ __stp_time_cpufreq_callback(struct notifier_block *self,
         case CPUFREQ_RESUMECHANGE:
             freqs = (struct cpufreq_freqs *)vfreqs;
             freq_khz = freqs->new;
-	    time = per_cpu_ptr(stp_time, smp_processor_id());
+            time = per_cpu_ptr(stp_time, freqs->cpu);
             write_seqlock_irqsave(&time->lock, flags);
-            time->cpufreq = freq_khz;
+            time->freq = freq_khz;
             write_sequnlock_irqrestore(&time->lock, flags);
             break;
     }
@@ -150,60 +185,80 @@ __stp_time_cpufreq_callback(struct notifier_block *self,
 struct notifier_block __stp_time_notifier = {
     .notifier_call = __stp_time_cpufreq_callback,
 };
+
+static int
+__stp_constant_freq(void)
+{
+#ifdef STAPCONF_CONSTANT_TSC
+    // If the CPU has constant tsc, we don't need to use cpufreq.
+    return boot_cpu_has(X86_FEATURE_CONSTANT_TSC);
+#elif defined (__ia64__) || defined (__s390__) || defined (__s390x__) || defined (__arm__)
+    // these architectures have constant time counter.
+    return 1;
+#else
+    return 0;
+#endif
+}
 #endif /* CONFIG_CPU_FREQ */
 
+/* This function is called during module unloading. */
 void
 _stp_kill_time(void)
 {
-	if (stp_time) {
-		int cpu;
-		for_each_online_cpu(cpu) {
-			stp_time_t *time = per_cpu_ptr(stp_time, cpu);
-			del_timer_sync(&time->timer);
-		}
+    if (stp_time) {
+        int cpu;
+        stp_timer_reregister = 0;
+        for_each_online_cpu(cpu) {
+            stp_time_t *time = per_cpu_ptr(stp_time, cpu);
+            del_timer_sync(&time->timer);
+        }
 #ifdef CONFIG_CPU_FREQ
-		cpufreq_unregister_notifier(&__stp_time_notifier,
-					    CPUFREQ_TRANSITION_NOTIFIER);
+        if (!__stp_constant_freq()) {
+            cpufreq_unregister_notifier(&__stp_time_notifier,
+                                        CPUFREQ_TRANSITION_NOTIFIER);
+        }
 #endif
-		
-		free_percpu(stp_time);
-	}
+
+        free_percpu(stp_time);
+    }
 }
 
+/* This function is called during module loading. */
 int
 _stp_init_time(void)
 {
     int ret = 0;
-    int cpu, freq_khz;
-    unsigned long flags;
 
     stp_time = alloc_percpu(stp_time_t);
     if (unlikely(stp_time == 0))
 	    return -1;
+    _stp_allocated_memory += sizeof(stp_time_t) * num_online_cpus();
     
+    stp_timer_reregister = 1;
     ret = on_each_cpu(__stp_init_time, NULL, 0, 1);
 
 #ifdef CONFIG_CPU_FREQ
-    if (ret) goto end;
+    if (!ret && !__stp_constant_freq()) {
+        ret = cpufreq_register_notifier(&__stp_time_notifier,
+                CPUFREQ_TRANSITION_NOTIFIER);
 
-    ret = cpufreq_register_notifier(&__stp_time_notifier,
-            CPUFREQ_TRANSITION_NOTIFIER);
-    if (ret) goto end;
-
-    for_each_online_cpu(cpu) {
-        preempt_disable();
-        freq_khz = cpufreq_get(cpu);
-        if (freq_khz > 0) {
-            stp_time_t *time = per_cpu_ptr(stp_time, cpu);
-            write_seqlock_irqsave(&time->lock, flags);
-            time->cpufreq = freq_khz;
-            write_sequnlock_irqrestore(&time->lock, flags);
+        if (!ret) {
+            int cpu;
+            for_each_online_cpu(cpu) {
+                unsigned long flags;
+                int freq_khz = cpufreq_get(cpu);
+                if (freq_khz > 0) {
+                    stp_time_t *time = per_cpu_ptr(stp_time, cpu);
+                    write_seqlock_irqsave(&time->lock, flags);
+                    time->freq = freq_khz;
+                    write_sequnlock_irqrestore(&time->lock, flags);
+                }
+            }
         }
-        preempt_enable();
     }
-end:
 #endif
-
+   if (ret)
+        _stp_kill_time();
     return ret;
 }
 
@@ -219,10 +274,11 @@ _stp_gettimeofday_ns(void)
 
     preempt_disable();
     time = per_cpu_ptr(stp_time, smp_processor_id());
+
     seq = read_seqbegin(&time->lock);
     base = time->base_ns;
     last = time->base_cycles;
-    freq = time->cpufreq;
+    freq = time->freq;
     while (unlikely(read_seqretry(&time->lock, seq))) {
         if (unlikely(++i >= MAXTRYLOCK))
             return 0;
@@ -230,12 +286,12 @@ _stp_gettimeofday_ns(void)
         seq = read_seqbegin(&time->lock);
         base = time->base_ns;
         last = time->base_cycles;
-        freq = time->cpufreq;
+        freq = time->freq;
     }
 
     delta = get_cycles() - last;
 
-    preempt_enable();
+    preempt_enable_no_resched();
 
 #if defined (__s390__) || defined (__s390x__)
     // The TOD clock on the s390 (read by get_cycles() ) 
@@ -244,6 +300,12 @@ _stp_gettimeofday_ns(void)
 
     delta = (delta * 125) >> 7;
 
+#elif defined (__arm__)
+
+    /* arm always returns 0 for get_cycles() */
+    /* so this is just a fake value until we get a real fix. */
+    delta = 1000;
+
 #else /* __s390__ || __s390x__ */
 
     // Verify units:
