@@ -7,12 +7,14 @@
  * redistribute it and/or modify it under the terms of the GNU General
  * Public License (GPL); either version 2, or (at your option) any
  * later version.
+ *
+ * The u32_swap(), generic_swap(), and sort() functions were adapted from
+ * lib/sort.c of kernel 2.6.22-rc5. It was written by Matt Mackall.
  */
 
 #ifndef _SYMBOLS_C_
 #define _SYMBOLS_C_
 #include "../sym.h"
-#include <linux/sort.h>
 
 spinlock_t _stp_module_lock = SPIN_LOCK_UNLOCKED;
 #define STP_TRYLOCK_MODULES  ({						\
@@ -102,6 +104,26 @@ static struct _stp_module * _stp_alloc_module_from_module (struct module *m)
 	return _stp_alloc_module(num, datasize);
 }
 
+static void _stp_free_module(struct _stp_module *mod)
+{
+	/* free symbol memory */
+	if (mod->num_symbols) {
+		if (mod->allocated & 1)
+			vfree(mod->symbols);
+		else
+			kfree(mod->symbols);
+		if (mod->allocated & 2)
+			vfree(mod->symbol_data);
+		else
+			kfree(mod->symbol_data);
+	}
+	if (mod->sections)
+		kfree(mod->sections);
+
+	/* free module memory */
+	kfree(mod);
+}
+
 /* Delete a module and free its memory. */
 /* The lock should already be held before calling this. */
 static void _stp_del_module(struct _stp_module *mod)
@@ -132,23 +154,8 @@ static void _stp_del_module(struct _stp_module *mod)
 		_stp_modules_by_addr[i] = _stp_modules_by_addr[i+1];
 
 	_stp_num_modules--;
-	
-	/* free symbol memory */
-	if (mod->num_symbols) {
-		if (mod->allocated & 1)
-			vfree(mod->symbols);
-		else
-			kfree(mod->symbols);
-		if (mod->allocated & 2)
-			vfree(mod->symbol_data);
-		else
-			kfree(mod->symbol_data);
-	}
-	if (mod->sections)
-		kfree(mod->sections);
 
-	/* free module memory */
-	kfree(mod);
+	_stp_free_module(mod);
 }
 
 static void _stp_free_modules(void)
@@ -208,6 +215,7 @@ static int _stp_do_symbols(const char __user *buf, int count)
 		_stp_symbol_state = 3;
                 /* NB: this mapping is used by kernel/_stext pseudo-relocations. */
 		_stp_modules[0]->text = _stp_kallsyms_lookup_name("_stext");
+		_stp_modules[0]->data = _stp_kallsyms_lookup_name("_etext");
 		_stp_modules_by_addr[0] = _stp_modules[0];
 		break;
 	default:
@@ -237,6 +245,75 @@ static void _stp_swap_symbol(void *x, void *y, int size)
 	b->symbol = symbol;
 }
 
+static void u32_swap(void *a, void *b, int size)
+{
+	u32 t = *(u32 *)a;
+	*(u32 *)a = *(u32 *)b;
+	*(u32 *)b = t;
+}
+
+static void generic_swap(void *a, void *b, int size)
+{
+	char t;
+
+	do {
+		t = *(char *)a;
+		*(char *)a++ = *(char *)b;
+		*(char *)b++ = t;
+	} while (--size > 0);
+}
+
+/**
+ * sort - sort an array of elements
+ * @base: pointer to data to sort
+ * @num: number of elements
+ * @size: size of each element
+ * @cmp: pointer to comparison function
+ * @swap: pointer to swap function or NULL
+ *
+ * This function does a heapsort on the given array. You may provide a
+ * swap function optimized to your element type.
+ *
+ * Sorting time is O(n log n) both on average and worst-case. While
+ * qsort is about 20% faster on average, it suffers from exploitable
+ * O(n*n) worst-case behavior and extra memory requirements that make
+ * it less suitable for kernel use.
+*/
+void _stp_sort(void *base, size_t num, size_t size,
+	int (*cmp)(const void *, const void *),
+	void (*swap)(void *, void *, int size))
+{
+	/* pre-scale counters for performance */
+	int i = (num/2 - 1) * size, n = num * size, c, r;
+
+	if (!swap)
+		swap = (size == 4 ? u32_swap : generic_swap);
+
+	/* heapify */
+	for ( ; i >= 0; i -= size) {
+		for (r = i; r * 2 + size < n; r  = c) {
+			c = r * 2 + size;
+			if (c < n - size && cmp(base + c, base + c + size) < 0)
+				c += size;
+			if (cmp(base + r, base + c) >= 0)
+				break;
+			swap(base + r, base + c, size);
+		}
+	}
+
+	/* sort */
+	for (i = n - size; i >= 0; i -= size) {
+		swap(base, base + i, size);
+		for (r = 0; r * 2 + size < i; r = c) {
+			c = r * 2 + size;
+			if (c < i - size && cmp(base + c, base + c + size) < 0)
+				c += size;
+			if (cmp(base + r, base + c) >= 0)
+				break;
+			swap(base + r, base + c, size);
+		}
+	}
+}
 
 /* Create a new _stp_module and load the symbols */
 static struct _stp_module *_stp_load_module_symbols (struct _stp_module *imod)
@@ -281,7 +358,7 @@ static struct _stp_module *_stp_load_module_symbols (struct _stp_module *imod)
 		module_put(m);
 
 		/* sort symbols by address */
-		sort (mod->symbols, num, sizeof(struct _stp_symbol), _stp_compare_addr, _stp_swap_symbol);
+		_stp_sort (mod->symbols, num, sizeof(struct _stp_symbol), _stp_compare_addr, _stp_swap_symbol);
 	}
 	return mod;
 }
@@ -306,15 +383,22 @@ static int _stp_module_exists(struct _stp_module *mod)
 	return 0;
 }
 
-static void _stp_ins_module(struct _stp_module *mod)
+static int _stp_ins_module(struct _stp_module *mod)
 {
-	int i, num, res;
+	int i, num, res, ret = 0;
 	unsigned long flags;
 
 	// kbug("insert %s\n", mod->name);
 
 	STP_LOCK_MODULES;
 
+	/* check for overflow */
+	if (_stp_num_modules == STP_MAX_MODULES) {
+		errk("Exceeded the limit of %d modules\n", STP_MAX_MODULES);
+		ret = -ENOMEM;
+		goto done;
+	}
+	
 	/* insert alphabetically in _stp_modules[] */
 	for (num = 1; num < _stp_num_modules; num++) {
 		res = strcmp(_stp_modules[num]->name, mod->name);
@@ -340,7 +424,9 @@ static void _stp_ins_module(struct _stp_module *mod)
 
 	_stp_num_modules++;
 
+done:
 	STP_UNLOCK_MODULES;
+	return ret;
 }
 
 
@@ -390,7 +476,10 @@ static int _stp_do_module(const char __user *buf, int count)
 		return 0;
 	}
 
-	_stp_ins_module(mod);
+	if (_stp_ins_module(mod) < 0) {
+		_stp_free_module(mod);
+		return -ENOMEM;
+	}
 	
 	return count;
 }
